@@ -1,4 +1,4 @@
-# Author: E. Noel Dawe <Noel.Dawe@cern.ch>
+# Author: Noel Dawe <Noel.Dawe@cern.ch>
 
 """
 This module provides the main GRL class and utility functions
@@ -18,17 +18,43 @@ try:
 except ImportError:
     USE_YAML = False
 
+import os
 import copy
 import urllib2
 from pprint import pprint
 from operator import sub, or_, and_, xor, itemgetter
 from .sorteddict import SortedDict
 import bisect
+import datetime
+import cStringIO
+import re
+
+
+__all__ = [
+    'clipped',
+    'diffed',
+    'ored',
+    'anded',
+    'xored',
+    'LumiblockRange',
+    'GRL',
+]
 
 
 def clipped(grl, startrun=None, startlb=None, endrun=None, endlb=None):
     """
-    Clip a GRL between startrun, startlb and endrun, endlb (inclusive)
+    Return a clipped GRL between startrun, startlb and
+    endrun, endlb (inclusive).
+
+    *grl*: GRL
+
+    *startrun*: [ int | None ]
+
+    *startlb*: [ int | None ]
+
+    *endrun*: [ int | None ]
+
+    *endlb*: [ int | None ]
     """
     grl_copy = copy.deepcopy(grl)
     grl_copy.clip(startrun=startrun, startlb=startlb,
@@ -39,6 +65,8 @@ def clipped(grl, startrun=None, startlb=None, endrun=None, endlb=None):
 def diffed(*args):
     """
     Return the difference of multiple GRLs: (((A-B)-C)-D)...)
+
+    *args*: tuple of GRLs
     """
     return reduce(sub, args)
 
@@ -46,6 +74,8 @@ def diffed(*args):
 def ored(*args):
     """
     Return the OR of multiple GRLs: A | B | C...
+
+    *args*: tuple of GRLs
     """
     return reduce(or_, args)
 
@@ -53,6 +83,8 @@ def ored(*args):
 def anded(*args):
     """
     Return the AND of multiple GRLs: A & B & C...
+
+    *args*: tuple of GRLs
     """
     return reduce(and_, args)
 
@@ -60,78 +92,73 @@ def anded(*args):
 def xored(*args):
     """
     Return the XOR of multiple GRLs: A ^ B ^ C...
+
+    *args*: tuple of GRLs
     """
     return reduce(xor, args)
 
 
-def _lbrange_intersects(lbrange1, lbrange2):
-    """
-    Determine if two lumiblock ranges intersect
-    """
-    if lbrange1[0] >= lbrange2[0] and lbrange1[0] <= lbrange2[1]:
-        return True
-    if lbrange1[1] >= lbrange2[0] and lbrange1[1] <= lbrange2[1]:
-        return True
-    if lbrange2[0] >= lbrange1[0] and lbrange2[0] <= lbrange1[1]:
-        return True
-    if lbrange2[1] >= lbrange1[0] and lbrange2[1] <= lbrange1[1]:
-        return True
-    return False
-
-
-def _lbrange_as_set(lbrange):
-    """
-    Convert lumiblock range to set of lumiblocks
-    """
-    return set(range(lbrange[0], lbrange[1] + 1))
-
-
-def _dict_to_grl(d):
-    """
-    Convert tuples to LumiblockRanges
-    """
-    o = {}
-    for run, lbranges in d.items():
-        o[run] = [LumiblockRange(a) for a in lbranges]
-    return o
-
-
-def _grl_to_dict(g):
-    """
-    Convert tuples to LumiblockRanges
-    """
-    o = {}
-    for run in g.iterruns():
-        lbranges = g[run]
-        o[run] = [(a[0], a[1]) for a in lbranges]
-    return o
-
-
 class LumiblockRange(tuple):
     """
     A 2-tuple consisting of the lower and upper
     bounds of a lumiblock range
     """
-    def __new__(cls, args):
-
+    def __new__(cls, *args):
+        """
+        *args*: [ tuple | list ]
+            2-tuple/list of lumiblock numbers in ascending order
+        """
+        if len(args) == 1 and isinstance(args[0], tuple):
+            args = args[0]
         if len(args) != 2:
             raise ValueError('lbrange must contain exactly 2 elements: %s' % \
                              (args,))
         for lumiblock in args:
-            if not isinstance(lumiblock, int):
-                raise TypeError('lbrange must contain integers only')
+            if not isinstance(lumiblock, (int, long)):
+                raise TypeError('lbrange must contain integers or longs only')
         if args[0] > args[1]:
             raise ValueError('lbrange in wrong order: %s' % (args,))
-
         return super(LumiblockRange, cls).__new__(cls, args)
 
-    def __cmp__(self, lumiblock):
+    def __contains__(self, lbn):
+        """
+        Determine whether a lumiblock is contained by this
+        lumiblock range
 
-        if lumiblock < self[0]:
-            return 1
-        if lumiblock > self[1]:
-            return -1
-        return 0
+        *lbn*: int
+        """
+        return self[0] <= lbn <= self[1]
+
+    def __cmp__(self, other):
+        """
+        Determine whether this lumiblock range should be placed
+        to the right or left of another lumiblock number or range
+
+        *other*: LumiblockRange
+        """
+        if isinstance(other, (int, long)):
+            if other in self:
+                return 0
+            return cmp(self[0], other)
+        return super(LumiblockRange, self).__cmp__(other)
+
+    def intersects(self, lbrange):
+        """
+        Determine if self intersects with another lumiblock range
+
+        *lbrange*: LumiblockRange
+        """
+        if lbrange[0] in self or lbrange[1] in self:
+            return True
+        if self[0] in lbrange or self[1] in lbrange:
+            return True
+        return False
+
+    def as_set(self):
+        """
+        Convert self to set of lumiblocks
+        """
+        return set(range(self[0], self[1] + 1))
 
 
 class GRL(object):
@@ -139,47 +166,158 @@ class GRL(object):
     The main GRL class holds a python dictionary
     mapping runs to a list of lumiblock ranges (2-tuples)
     """
-    def __init__(self, grl=None):
+    formats = [
+        'xml',
+        'yml',
+        'txt',
+        'py',
+        'cut'
+    ]
+
+    def __init__(self, grl=None, from_string=False):
         """
-        grl may be a file name or URL of a valid GRL file, or None
+        *grl*: [ dict | str | None ]
+
+        *from_string*: bool
+            If True, interpret grl as xml string and not filename
         """
+        self.name = 'GRL'
+        self.version = '1.0'
+        self.metadata = []
         self.__grl = SortedDict()
         if not grl:
             return
         if isinstance(grl, dict):
-            self.__grl = SortedDict(_dict_to_grl(grl))
+            self.from_dict(grl)
             return
-        if type(grl) in [str, file]:
+        if isinstance(grl, basestring) and from_string:
+            self.from_string(grl)
+            return
+        elif from_string:
+            raise TypeError("grl is non-string type %s while "
+                            "using from_string" % type(grl))
+        elif isinstance(grl, (basestring, file)):
             filename = grl
-            if type(grl) is str:
+            if isinstance(grl, basestring):
+                # is grl a URL?
                 if grl.startswith("http://"):
                     grl = urllib2.urlopen(grl)
-            elif type(grl) is file:
+                # is grl a ROOT file path?
+                elif re.search('.root:/', grl):
+                    # one place where goodruns requires ROOT
+                    try:
+                        import ROOT
+                        ROOT.PyConfig.IgnoreCommandLineOptions = True
+                    except ImportError:
+                        raise ImportError('Specified GRL in ROOT file '
+                                          'but cannot import ROOT. Are ROOT '
+                                          'and PyROOT installed?')
+                    cwd = ROOT.gDirectory
+                    filename, _, path = grl.partition(':/')
+                    root_file = ROOT.TFile.Open(filename)
+                    if not root_file:
+                        raise IOError('Could not open ROOT file: %s' %
+                                      filename)
+                    grl = root_file.Get(path)
+                    if not grl:
+                        raise ValueError('Path %s does not exist in '
+                                         'ROOT file %s' % (path, filename))
+                    if not isinstance(grl, ROOT.TObjString):
+                        raise TypeError('Object at %s is not a '
+                                        'ROOT.TObjString' % path)
+                    self.from_string(str(grl.GetString()))
+                    root_file.Close()
+                    # return to previous directory
+                    cwd.cd()
+                    for run in self.iterruns():
+                        self.__grl[run].sort()
+                        self.__optimize(run)
+                    return
+            elif isinstance(grl, file):
                 filename = grl.name
-            if filename == "<stdin>" or filename.endswith('.xml') or \
-                    filename.startswith("http://"):
+            name, ext = os.path.splitext(filename)
+            if filename == "<stdin>" or ext == '.xml':
                 tree = ET.parse(grl)
-                lbcols = tree.getroot().findall(
-                    'NamedLumiRange/LumiBlockCollection')
-                for lbcol in lbcols:
-                    run = int(lbcol.find('Run').text)
-                    lbs = lbcol.findall('LBRange')
-                    for lumiblock in lbs:
-                        self.insert(run,
-                            LumiblockRange((int(lumiblock.attrib['Start']),
-                                            int(lumiblock.attrib['End']))))
-            elif filename.endswith('.yml'):
+                self.from_xml(tree)
+            elif ext == '.yml':
                 if USE_YAML:
-                    self.__grl = SortedDict(_dict_to_grl(grl))
+                    if isinstance(grl, file):
+                        self.from_dict(yaml.load(grl))
+                    else:
+                        with open(grl) as grl_file:
+                            self.from_dict(yaml.load(grl_file))
                 else:
                     raise ImportError("PyYAML module not found")
             else:
                 raise ValueError(
-                    "File %s is not recognized as a valid GRL format"
-                    % filename)
+                    "File %s does not have valid GRL extension: %s"
+                    % (filename, ext))
             for run in self.iterruns():
                 self.__grl[run].sort()
                 self.__optimize(run)
+            return
+        raise TypeError("Unable to initialize GRL from a %s" % type(grl))
+
+    def from_string(self, string):
+        """
+        Insert runs and lumiblocks from XML string
+
+        *string*: str
+        """
+        tree = ET.fromstring(string)
+        self.from_xml(tree)
+
+    def from_xml(self, tree):
+        """
+        Insert runs and lumiblocks from XML
+
+        *tree*: ElementTree
+        """
+        name = tree.find('NamedLumiRange/Name')
+        if name is not None:
+            self.name = name.text
+        version = tree.find('NamedLumiRange/Version')
+        if version is not None:
+            self.version = version.text
+        metadata = tree.findall('NamedLumiRange/Metadata')
+        if metadata is not None:
+            self.metadata = metadata
+        lbcols = tree.findall(
+            'NamedLumiRange/LumiBlockCollection')
+        if lbcols is None:
+            return
+        for lbcol in lbcols:
+            run = int(lbcol.find('Run').text)
+            lbs = lbcol.findall('LBRange')
+            for lumiblock in lbs:
+                self.insert(run,
+                    LumiblockRange(int(lumiblock.attrib['Start']),
+                                   int(lumiblock.attrib['End'])))
+
+    def from_dict(self, d):
+        """
+        Convert dict to GRL
+
+        *d*: dict
+        """
+        o = {}
+        for run, lbranges in d.items():
+            o[run] = [LumiblockRange(*a) for a in lbranges]
+        self.__grl.update(o)
+
+    def to_dict(self):
+        """
+        Convert self to dict
+        """
+        o = {}
+        for run, lbranges in self.items():
+            o[run] = [(a[0], a[1]) for a in lbranges]
+        return o
+
+    def __merge_metadata(self, other=None):
+
+        # drop metadata for now
+        self.metadata = []
 
     def __copy__(self):
 
@@ -216,25 +354,32 @@ class GRL(object):
     def __getitem__(self, run):
         """
         Return list of lumiblock ranges for a run
+
+        *run*: int
         """
         return self.__grl[run]
 
     def __delitem__(self, run):
         """
         Remove run and associated lumiblock ranges from GRL
+
+        *run*: int
         """
         del self.__grl[run]
 
     def __contains__(self, runlb):
         """
-        Pass the tuple (run, lbn)
+        Returns True if this GRL contains a run and lumiblock
+
+        *runlb*: tuple
+            2-tuple of ints containing run number and lumiblock number
         """
         run, lbn = runlb
         if run in self.__grl:
             lbranges = self.__grl[run]
             # Locate the LumiblockRange containing lbn
             i = bisect.bisect_left(lbranges, lbn)
-            if i != len(lbranges) and lbranges[i] == lbn:
+            if (i != len(lbranges)) and (lbn in lbranges[i]):
                 return True
         return False
 
@@ -242,12 +387,18 @@ class GRL(object):
         """
         Iterate over runs in GRL
         """
-        return iter(self.iterruns())
+        return self.iterruns()
 
     def items(self):
         """
         Iterate over (run, lbranges) in GRL
         """
+        return self.__grl.items()
+
+    def iterlbranges(self):
+        """
+        Iterate over (run, lbrange) in GRL
+        """
         for run, lbranges in self.__grl.items():
             for lbrange in lbranges:
                 yield (run, lbrange)
@@ -268,20 +419,26 @@ class GRL(object):
     def has_run(self, run):
         """
         Returns True if run is in GRL, else False
+
+        *run*: int
         """
         return run in self.__grl
 
     def insert(self, run, lbrange):
         """
         Insert a lumiblock range into a run
+
+        *run*: int
+
+        *lbrange*: [ LumiblockRange | tuple ]
         """
         if not isinstance(run, int):
             raise TypeError('run must be an integer')
         if not isinstance(lbrange, LumiblockRange):
-            lbrange = LumiblockRange(lbrange)
+            lbrange = LumiblockRange(*lbrange)
         try:
             lbranges = self.__grl[run]
-            i = bisect.bisect(lbranges, lbrange[0])
+            i = bisect.bisect(lbranges, lbrange)
             lbranges.insert(i, lbrange)
             self.__optimize(run)
         except KeyError:
@@ -290,10 +447,14 @@ class GRL(object):
     def remove(self, run, lbrange):
         """
         Remove a lumiblock range from a run
+
+        *run*: int
+
+        *lbrange*: LumiblockRange
         """
         if run in self.__grl:
             if not isinstance(lbrange, LumiblockRange):
-                lbrange = LumiblockRange(lbrange)
+                lbrange = LumiblockRange(*lbrange)
             lbranges = self.__grl[run]
             for mylbrange in lbranges[:]:
                 if lbrange[1] < mylbrange[0]:
@@ -303,23 +464,22 @@ class GRL(object):
                     break
                 elif lbrange[0] > mylbrange[0] and lbrange[1] < mylbrange[1]:
                     # embedded: must split
-                    left_lbrange = LumiblockRange((mylbrange[0],
-                                                   lbrange[0] - 1))
-                    right_lbrange = LumiblockRange((lbrange[1] + 1,
-                                                    mylbrange[1]))
+                    left_lbrange = LumiblockRange(mylbrange[0],
+                                                  lbrange[0] - 1)
+                    right_lbrange = LumiblockRange(lbrange[1] + 1,
+                                                   mylbrange[1])
                     index = lbranges.index(mylbrange)
                     lbranges[index] = left_lbrange
                     lbranges.insert(index + 1, right_lbrange)
                     break
-                elif _lbrange_intersects(lbrange, mylbrange):
-                    diff = _lbrange_as_set(mylbrange).difference(
-                        _lbrange_as_set(lbrange))
+                elif lbrange.intersects(mylbrange):
+                    diff = mylbrange.as_set().difference(lbrange.as_set())
                     if not diff:  # empty set
                         lbranges.remove(mylbrange)
                         if len(lbranges) == 0:
                             break
                         continue
-                    newlbrange = LumiblockRange((min(diff), max(diff)))
+                    newlbrange = LumiblockRange(min(diff), max(diff))
                     lbranges[lbranges.index(mylbrange)] = newlbrange
                 elif mylbrange[0] > lbrange[1]:
                     break
@@ -329,6 +489,14 @@ class GRL(object):
     def clip(self, startrun=None, startlb=None, endrun=None, endlb=None):
         """
         Clip the GRL between startrun, startlb and endrun, endlb (inclusive)
+
+        *startrun*: [ int | None ]
+
+        *startlb*: [ int | None ]
+
+        *endrun*: [ int | None ]
+
+        *endlb*: [ int | None ]
         """
         for run in self.runs():
             if startrun is not None:
@@ -344,7 +512,7 @@ class GRL(object):
                                  startlb <= lbrange[1]:
                                 self.__grl[run][
                                     self.__grl[run].index(lbrange)
-                                    ] = LumiblockRange((startlb, lbrange[1]))
+                                    ] = LumiblockRange(startlb, lbrange[1])
                         if len(self.__grl[run]) == 0:
                             del self[run]
             if endrun is not None:
@@ -359,13 +527,15 @@ class GRL(object):
                             elif endlb >= lbrange[0] and endlb <= lbrange[1]:
                                 self[run][
                                     self[run].index(lbrange)
-                                    ] = LumiblockRange((lbrange[0], endlb))
+                                    ] = LumiblockRange(lbrange[0], endlb)
                         if len(self[run]) == 0:
                             del self[run]
 
     def __optimize(self, run):
         """
         Merge lumiblock ranges
+
+        *run*: int
         """
         lbranges = self.__grl[run]
         if len(lbranges) == 0:
@@ -383,8 +553,8 @@ class GRL(object):
                     merged = True
                     break
                 elif lbranges[first][1] + 1 >= lbranges[_next][0]:
-                    lbranges[first] = LumiblockRange((lbranges[first][0],
-                                                      lbranges[_next][1]))
+                    lbranges[first] = LumiblockRange(lbranges[first][0],
+                                                     lbranges[_next][1])
                     for index in xrange(first + 1, _next + 1):
                         lbranges.pop(_next)
                     merged = True
@@ -394,6 +564,14 @@ class GRL(object):
             if not merged:
                 first += 1
 
+    def __eq__(self, other):
+
+        return self.__grl == other.__grl
+
+    def __ne__(self, other):
+
+        return not self.__eq__(other)
+
     def __add__(self, other):
 
         grlcopy = copy.deepcopy(self)
@@ -402,8 +580,11 @@ class GRL(object):
 
     def __iadd__(self, other):
 
-        for run, lbrange in other.items():
-            self.insert(run, lbrange)
+        if isinstance(other, basestring):
+            other = GRL(other, from_string=True)
+        for run, lbrange in other.iterlbranges():
+            self.insert(run, LumiblockRange(lbrange))
+        self.__merge_metadata(other)
         return self
 
     def __sub__(self, other):
@@ -414,26 +595,50 @@ class GRL(object):
 
     def __isub__(self, other):
 
-        for run, lbrange in other.items():
+        if isinstance(other, basestring):
+            other = GRL(other, from_string=True)
+        for run, lbrange in other.iterlbranges():
             self.remove(run, lbrange)
+        self.__merge_metadata(other)
         return self
 
     def __and__(self, other):
 
         return self - (self - other)
 
+    def __iand__(self, other):
+
+        self -= (self - other)
+        return self
+
     def __or__(self, other):
 
         return self + other
 
+    def __ior__(self, other):
+
+        self += other
+        return self
+
     def __xor__(self, other):
 
         return (self | other) - (self & other)
 
+    def __ixor__(self, other):
+
+        grlcopy = copy.deepcopy(self)
+        self |= other
+        self -= (grlcopy & other)
+        return self
+
     def cut(self, runname='RunNumber', lbname='lbn'):
         """
         Convert this GRL into a TCut expression.
         This method is really meant to be a joke and should be of no use.
+
+        *runname*: str
+
+        *lbname*: str
         """
         cut = ''
         for run in self.iterruns():
@@ -451,21 +656,76 @@ class GRL(object):
                 cut = newcut
         return cut
 
-    formats = [
-        'xml',
-        'yml',
-        'txt',
-        'py',
-        'cut'
-    ]
+    def str(self, format='xml'):
+        """
+        Return string repr of self in the specified format
+
+        *format*: str
+        """
+        str_io = cStringIO.StringIO()
+        self.write(filehandle=str_io, format=format)
+        return str_io.getvalue()
+
+    def save(self, name):
+        """
+        Save GRL to file by name. Determine format from
+        extension.
+
+        *name*: str
+        """
+        # is name a ROOT file path?
+        if re.search('.root:/', name):
+            # one place where goodruns requires ROOT
+            try:
+                import ROOT
+                ROOT.PyConfig.IgnoreCommandLineOptions = True
+            except ImportError:
+                raise ImportError('Attempting to save GRL in ROOT file '
+                                  'but cannot import ROOT. Are ROOT and PyROOT '
+                                  'installed?')
+            cwd = ROOT.gDirectory
+            filename, _, path = name.partition(':/')
+            root_file = ROOT.TFile.Open(filename, 'UPDATE')
+            if not root_file:
+                raise IOError('Could not open ROOT file: %s' %
+                              filename)
+            head, tail = os.path.split(os.path.normpath(path))
+            if head and not root_file.cd(head):
+                raise ValueError('Path %s does not exist in file %s' %
+                                 (head, filename))
+            xml_string = ROOT.TObjString(self.str())
+            xml_string.Write(tail)
+            root_file.Close()
+            # return to previous directory
+            cwd.cd()
+        else:
+            _, ext = os.path.splitext(name)
+            # ignore period
+            ext = ext[1:]
+            if ext not in GRL.formats:
+                raise ValueError("Filename %s does not have "
+                                 "a valid GRL extension." %
+                                 name)
+            with open(name, 'w') as filehandle:
+                self.write(filehandle, format=ext)
 
     def write(self, filehandle, format='xml'):
         """
-        Write the GRL in a specified format to the file object.
+        Write the GRL in the specified format to the file object.
+
+        *filehandle*: file
+
+        *format*: str
         """
         if format == 'xml':
             root = ET.Element('LumiRangeCollection')
             subroot = ET.SubElement(root, 'NamedLumiRange')
+            name = ET.SubElement(subroot, 'Name')
+            name.text = self.name
+            version = ET.SubElement(subroot, 'Version')
+            version.text = self.version
+            for meta in self.metadata:
+                subroot.append(meta)
             for run in self.iterruns():
                 lumiblocks = self.__grl[run]
                 lbcol = ET.SubElement(subroot, 'LumiBlockCollection')
@@ -475,15 +735,24 @@ class GRL(object):
                     lbrange = ET.SubElement(lbcol, 'LBRange')
                     lbrange.set('Start', str(lumiblock[0]))
                     lbrange.set('End', str(lumiblock[1]))
+            date = datetime.datetime.now().strftime("%Y-%m-%d at %H:%M:%S")
+            meta = (
+            '''<!DOCTYPE LumiRangeCollection SYSTEM '''
+            '''"http://atlas-runquery.cern.ch/LumiRangeCollection.dtd">\n'''
+            '''<!-- This document was created by goodruns: '''
+            '''http://pypi.python.org/pypi/goodruns/ on %s -->\n''' % date)
             tree = ET.ElementTree(root)
             if USE_LXML:
+                filehandle.write('<?xml version="1.0"?>\n')
+                filehandle.write(meta)
                 tree.write(filehandle, pretty_print=True)
             else:
                 # get pretty XML from ElementTree
-                xml = minidom.parseString(ET.tostring(tree.getroot()))
+                xml = minidom.parseString(meta +
+                                          ET.tostring(tree.getroot(), 'utf-8'))
                 filehandle.write(xml.toprettyxml(indent='   '))
         elif format in ('yml', 'yaml'):
-            filehandle.write(yaml.dump(_grl_to_dict(self)))
+            filehandle.write(yaml.dump(self.to_dict()))
         elif format == 'txt':
             filehandle.write(str(self) + '\n')
         elif format in ('py', 'python'):
